@@ -1,6 +1,6 @@
 /**
  * Object oriented C module to send ICMP and ICMPv6 `echo's.
- * Copyright (C) 2006  Florian octo Forster <octo at verplant.org>
+ * Copyright (C) 2006-2010  Florian octo Forster <octo at verplant.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -62,6 +62,15 @@
 #include <sys/types.h>
 #endif
 
+#if USE_NCURSES
+# define NCURSES_OPAQUE 1
+# include <ncurses.h>
+
+# define OPING_GREEN 1
+# define OPING_YELLOW 2
+# define OPING_RED 3
+#endif
+
 #include "oping.h"
 
 #ifndef _POSIX_SAVED_IDS
@@ -73,6 +82,7 @@ typedef struct ping_context
 	char host[NI_MAXHOST];
 	char addr[NI_MAXHOST];
 
+	int index;
 	int req_sent;
 	int req_rcvd;
 
@@ -80,6 +90,10 @@ typedef struct ping_context
 	double latency_max;
 	double latency_total;
 	double latency_total_square;
+
+#if USE_NCURSES
+	WINDOW *window;
+#endif
 } ping_context_t;
 
 static double  opt_interval   = 1.0;
@@ -90,15 +104,21 @@ static char   *opt_filename   = NULL;
 static int     opt_count      = -1;
 static int     opt_send_ttl   = 64;
 
-static void sigint_handler (int signal)
+static int host_num = 0;
+
+#if USE_NCURSES
+static WINDOW *main_win = NULL;
+#endif
+
+static void sigint_handler (int signal) /* {{{ */
 {
 	/* Make compiler happy */
 	signal = 0;
 	/* Exit the loop */
 	opt_count = 0;
-}
+} /* }}} void sigint_handler */
 
-static ping_context_t *context_create (void)
+static ping_context_t *context_create (void) /* {{{ */
 {
 	ping_context_t *ret;
 
@@ -112,15 +132,107 @@ static ping_context_t *context_create (void)
 	ret->latency_total = 0.0;
 	ret->latency_total_square = 0.0;
 
+#if USE_NCURSES
+	ret->window = NULL;
+#endif
+
 	return (ret);
-}
+} /* }}} ping_context_t *context_create */
 
-static void context_destroy (ping_context_t *context)
+static void context_destroy (ping_context_t *context) /* {{{ */
 {
-	free (context);
-}
+	if (context == NULL)
+		return;
 
-static void usage_exit (const char *name, int status)
+#if USE_NCURSES
+	if (context->window != NULL)
+	{
+		delwin (context->window);
+		context->window = NULL;
+	}
+#endif
+
+	free (context);
+} /* }}} void context_destroy */
+
+static double context_get_average (ping_context_t *ctx) /* {{{ */
+{
+	double num_total;
+
+	if (ctx == NULL)
+		return (-1.0);
+
+	if (ctx->req_rcvd < 1)
+		return (-0.0);
+
+	num_total = (double) ctx->req_rcvd;
+	return (ctx->latency_total / num_total);
+} /* }}} double context_get_average */
+
+static double context_get_stddev (ping_context_t *ctx) /* {{{ */
+{
+	double num_total;
+
+	if (ctx == NULL)
+		return (-1.0);
+
+	if (ctx->req_rcvd < 1)
+		return (-0.0);
+	else if (ctx->req_rcvd < 2)
+		return (0.0);
+
+	num_total = (double) ctx->req_rcvd;
+	return (sqrt (((num_total * ctx->latency_total_square)
+					- (ctx->latency_total * ctx->latency_total))
+				/ (num_total * (num_total - 1.0))));
+} /* }}} double context_get_stddev */
+
+static double context_get_packet_loss (const ping_context_t *ctx) /* {{{ */
+{
+	if (ctx == NULL)
+		return (-1.0);
+
+	if (ctx->req_sent < 1)
+		return (0.0);
+
+	return (100.0 * (ctx->req_sent - ctx->req_rcvd)
+			/ ((double) ctx->req_sent));
+} /* }}} double context_get_packet_loss */
+
+static int ping_initialize_contexts (pingobj_t *ping) /* {{{ */
+{
+	pingobj_iter_t *iter;
+	int index;
+
+	if (ping == NULL)
+		return (EINVAL);
+
+	index = 0;
+	for (iter = ping_iterator_get (ping);
+			iter != NULL;
+			iter = ping_iterator_next (iter))
+	{
+		ping_context_t *context;
+		size_t buffer_size;
+
+		context = context_create ();
+		context->index = index;
+
+		buffer_size = sizeof (context->host);
+		ping_iterator_get_info (iter, PING_INFO_HOSTNAME, context->host, &buffer_size);
+
+		buffer_size = sizeof (context->addr);
+		ping_iterator_get_info (iter, PING_INFO_ADDRESS, context->addr, &buffer_size);
+
+		ping_iterator_set_context (iter, (void *) context);
+
+		index++;
+	}
+
+	return (0);
+} /* }}} int ping_initialize_contexts */
+
+static void usage_exit (const char *name, int status) /* {{{ */
 {
 	int name_length;
 
@@ -143,9 +255,9 @@ static void usage_exit (const char *name, int status)
 			"for contributions see `AUTHORS'\n",
 			name);
 	exit (status);
-}
+} /* }}} void usage_exit */
 
-static int read_options (int argc, char **argv)
+static int read_options (int argc, char **argv) /* {{{ */
 {
 	int optchar;
 
@@ -227,9 +339,280 @@ static int read_options (int argc, char **argv)
 	}
 
 	return (optind);
-}
+} /* }}} read_options */
 
-static void print_host (pingobj_iter_t *iter)
+static void time_normalize (struct timespec *ts) /* {{{ */
+{
+	while (ts->tv_nsec < 0)
+	{
+		if (ts->tv_sec == 0)
+		{
+			ts->tv_nsec = 0;
+			return;
+		}
+
+		ts->tv_sec  -= 1;
+		ts->tv_nsec += 1000000000;
+	}
+
+	while (ts->tv_nsec >= 1000000000)
+	{
+		ts->tv_sec  += 1;
+		ts->tv_nsec -= 1000000000;
+	}
+} /* }}} void time_normalize */
+
+static void time_calc (struct timespec *ts_dest, /* {{{ */
+		const struct timespec *ts_int,
+		const struct timeval  *tv_begin,
+		const struct timeval  *tv_end)
+{
+	ts_dest->tv_sec = tv_begin->tv_sec + ts_int->tv_sec;
+	ts_dest->tv_nsec = (tv_begin->tv_usec * 1000) + ts_int->tv_nsec;
+	time_normalize (ts_dest);
+
+	/* Assure that `(begin + interval) > end'.
+	 * This may seem overly complicated, but `tv_sec' is of type `time_t'
+	 * which may be `unsigned. *sigh* */
+	if ((tv_end->tv_sec > ts_dest->tv_sec)
+			|| ((tv_end->tv_sec == ts_dest->tv_sec)
+				&& ((tv_end->tv_usec * 1000) > ts_dest->tv_nsec)))
+	{
+		ts_dest->tv_sec  = 0;
+		ts_dest->tv_nsec = 0;
+		return;
+	}
+
+	ts_dest->tv_sec = ts_dest->tv_sec - tv_end->tv_sec;
+	ts_dest->tv_nsec = ts_dest->tv_nsec - (tv_end->tv_usec * 1000);
+	time_normalize (ts_dest);
+} /* }}} void time_calc */
+
+#if USE_NCURSES
+static int update_stats_from_context (ping_context_t *ctx) /* {{{ */
+{
+	if ((ctx == NULL) || (ctx->window == NULL))
+		return (EINVAL);
+
+	werase (ctx->window);
+
+	box (ctx->window, 0, 0);
+	wattron (ctx->window, A_BOLD);
+	mvwprintw (ctx->window, /* y = */ 0, /* x = */ 5,
+			" %s ", ctx->host);
+	wattroff (ctx->window, A_BOLD);
+	wprintw (ctx->window, "ping statistics ");
+	mvwprintw (ctx->window, /* y = */ 1, /* x = */ 2,
+			"%i packets transmitted, %i received, %.2f%% packet "
+			"loss, time %.1fms",
+			ctx->req_sent, ctx->req_rcvd,
+			context_get_packet_loss (ctx),
+			ctx->latency_total);
+	if (ctx->req_rcvd != 0)
+	{
+		double average;
+		double deviation;
+
+		average = context_get_average (ctx);
+		deviation = context_get_stddev (ctx);
+			
+		mvwprintw (ctx->window, /* y = */ 2, /* x = */ 2,
+				"rtt min/avg/max/sdev = %.3f/%.3f/%.3f/%.3f ms",
+				ctx->latency_min,
+				average,
+				ctx->latency_max,
+				deviation);
+	}
+
+	wrefresh (ctx->window);
+
+	return (0);
+} /* }}} int update_stats_from_context */
+
+static int on_resize (pingobj_t *ping) /* {{{ */
+{
+	pingobj_iter_t *iter;
+	int width = 0;
+	int height = 0;
+	int main_win_height;
+
+	getmaxyx (stdscr, height, width);
+	if ((height < 1) || (width < 1))
+		return (EINVAL);
+
+	main_win_height = height - (4 * host_num);
+	wresize (main_win, main_win_height, /* width = */ width);
+	/* Allow scrolling */
+	scrollok (main_win, TRUE);
+	/* wsetscrreg (main_win, 0, main_win_height - 1); */
+	/* Allow hardware accelerated scrolling. */
+	idlok (main_win, TRUE);
+	wrefresh (main_win);
+
+	for (iter = ping_iterator_get (ping);
+			iter != NULL;
+			iter = ping_iterator_next (iter))
+	{
+		ping_context_t *context;
+
+		context = ping_iterator_get_context (iter);
+		if (context == NULL)
+			continue;
+
+		if (context->window != NULL)
+		{
+			delwin (context->window);
+			context->window = NULL;
+		}
+		context->window = newwin (/* height = */ 4,
+				/* width = */ 0,
+				/* y = */ main_win_height + (4 * context->index),
+				/* x = */ 0);
+	}
+
+	return (0);
+} /* }}} */
+
+static int check_resize (pingobj_t *ping) /* {{{ */
+{
+	int need_resize = 0;
+
+	while (42)
+	{
+		int key = wgetch (stdscr);
+		if (key == ERR)
+			break;
+		else if (key == KEY_RESIZE)
+			need_resize = 1;
+	}
+
+	if (need_resize)
+		return (on_resize (ping));
+	else
+		return (0);
+} /* }}} int check_resize */
+
+static int pre_loop_hook (pingobj_t *ping) /* {{{ */
+{
+	pingobj_iter_t *iter;
+	int width = 0;
+	int height = 0;
+	int main_win_height;
+
+	initscr ();
+	cbreak ();
+	noecho ();
+	nodelay (stdscr, TRUE);
+
+	getmaxyx (stdscr, height, width);
+	if ((height < 1) || (width < 1))
+		return (EINVAL);
+
+	if (has_colors () == TRUE)
+	{
+		start_color ();
+		init_pair (OPING_GREEN,  COLOR_GREEN,  /* default = */ 0);
+		init_pair (OPING_YELLOW, COLOR_YELLOW, /* default = */ 0);
+		init_pair (OPING_RED,    COLOR_RED,    /* default = */ 0);
+	}
+
+	main_win_height = height - (4 * host_num);
+	main_win = newwin (/* height = */ main_win_height,
+			/* width = */ 0,
+			/* y = */ 0, /* x = */ 0);
+	/* Allow scrolling */
+	scrollok (main_win, TRUE);
+	/* wsetscrreg (main_win, 0, main_win_height - 1); */
+	/* Allow hardware accelerated scrolling. */
+	idlok (main_win, TRUE);
+	wmove (main_win, /* y = */ main_win_height - 1, /* x = */ 0);
+	wrefresh (main_win);
+
+	for (iter = ping_iterator_get (ping);
+			iter != NULL;
+			iter = ping_iterator_next (iter))
+	{
+		ping_context_t *context;
+
+		context = ping_iterator_get_context (iter);
+		if (context == NULL)
+			continue;
+
+		if (context->window != NULL)
+		{
+			delwin (context->window);
+			context->window = NULL;
+		}
+		context->window = newwin (/* height = */ 4,
+				/* width = */ 0,
+				/* y = */ main_win_height + (4 * context->index),
+				/* x = */ 0);
+	}
+
+
+	/* Don't know what good this does exactly, but without this code
+	 * "check_resize" will be called right after startup and *somehow*
+	 * this leads to display errors. If we purge all initial characters
+	 * here, the problem goes away. "wgetch" is non-blocking due to
+	 * "nodelay" (see above). */
+	while (wgetch (stdscr) != ERR)
+	{
+		/* eat up characters */;
+	}
+
+	return (0);
+} /* }}} int pre_loop_hook */
+
+static int pre_sleep_hook (pingobj_t *ping) /* {{{ */
+{
+	return (check_resize (ping));
+} /* }}} int pre_sleep_hook */
+
+static int post_sleep_hook (pingobj_t *ping) /* {{{ */
+{
+	return (check_resize (ping));
+} /* }}} int pre_sleep_hook */
+#else /* if !USE_NCURSES */
+static int pre_loop_hook (pingobj_t *ping) /* {{{ */
+{
+	pingobj_iter_t *iter;
+
+	for (iter = ping_iterator_get (ping);
+			iter != NULL;
+			iter = ping_iterator_next (iter))
+	{
+		ping_context_t *ctx;
+		size_t buffer_size;
+
+		ctx = ping_iterator_get_context (iter);
+		if (ctx == NULL)
+			continue;
+
+		buffer_size = 0;
+		ping_iterator_get_info (iter, PING_INFO_DATA, NULL, &buffer_size);
+
+		printf ("PING %s (%s) %zu bytes of data.\n",
+				ctx->host, ctx->addr, buffer_size);
+	}
+
+	return (0);
+} /* }}} int pre_loop_hook */
+
+static int pre_sleep_hook (__attribute__((unused)) pingobj_t *ping) /* {{{ */
+{
+	fflush (stdout);
+
+	return (0);
+} /* }}} int pre_sleep_hook */
+
+static int post_sleep_hook (__attribute__((unused)) pingobj_t *ping) /* {{{ */
+{
+	return (0);
+} /* }}} int post_sleep_hook */
+#endif
+
+static void update_host_hook (pingobj_iter_t *iter, /* {{{ */
+		int index)
 {
 	double          latency;
 	unsigned int    sequence;
@@ -259,6 +642,12 @@ static void print_host (pingobj_iter_t *iter)
 
 	context = (ping_context_t *) ping_iterator_get_context (iter);
 
+#if USE_NCURSES
+# define HOST_PRINTF(...) wprintw(main_win, __VA_ARGS__)
+#else
+# define HOST_PRINTF(...) printf(__VA_ARGS__)
+#endif
+
 	context->req_sent++;
 	if (latency > 0.0)
 	{
@@ -271,67 +660,116 @@ static void print_host (pingobj_iter_t *iter)
 		if ((context->latency_min < 0.0) || (context->latency_min > latency))
 			context->latency_min = latency;
 
-		printf ("%zu bytes from %s (%s): icmp_seq=%u ttl=%i time=%.2f ms\n",
+#if USE_NCURSES
+		if (has_colors () == TRUE)
+		{
+			int color = OPING_GREEN;
+			double average = context_get_average (context);
+			double stddev = context_get_stddev (context);
+
+			if ((latency < (average - (2 * stddev)))
+					|| (latency > (average + (2 * stddev))))
+				color = OPING_RED;
+			else if ((latency < (average - stddev))
+					|| (latency > (average + stddev)))
+				color = OPING_YELLOW;
+
+			HOST_PRINTF ("%zu bytes from %s (%s): icmp_seq=%u ttl=%i "
+					"time=",
+					data_len, context->host, context->addr,
+					sequence, recv_ttl);
+			wattron (main_win, COLOR_PAIR(color));
+			HOST_PRINTF ("%.2f", latency);
+			wattroff (main_win, COLOR_PAIR(color));
+			HOST_PRINTF (" ms\n");
+		}
+		else
+		{
+#endif
+		HOST_PRINTF ("%zu bytes from %s (%s): icmp_seq=%u ttl=%i "
+				"time=%.2f ms\n",
 				data_len,
 				context->host, context->addr,
 				sequence, recv_ttl, latency);
+#if USE_NCURSES
+		}
+#endif
 	}
 	else
 	{
-		printf ("echo reply from %s (%s): icmp_seq=%u timeout\n",
+#if USE_NCURSES
+		if (has_colors () == TRUE)
+		{
+			HOST_PRINTF ("echo reply from %s (%s): icmp_seq=%u ",
+					context->host, context->addr,
+					sequence);
+			wattron (main_win, COLOR_PAIR(OPING_RED) | A_BOLD);
+			HOST_PRINTF ("timeout");
+			wattroff (main_win, COLOR_PAIR(OPING_RED) | A_BOLD);
+			HOST_PRINTF ("\n");
+		}
+		else
+		{
+#endif
+		HOST_PRINTF ("echo reply from %s (%s): icmp_seq=%u timeout\n",
 				context->host, context->addr,
 				sequence);
+#if USE_NCURSES
+		}
+#endif
 	}
-}
 
-static void time_normalize (struct timespec *ts)
+#if USE_NCURSES
+	update_stats_from_context (context);
+	wrefresh (main_win);
+#endif
+} /* }}} void update_host_hook */
+
+static int post_loop_hook (pingobj_t *ping) /* {{{ */
 {
-	while (ts->tv_nsec < 0)
+	pingobj_iter_t *iter;
+
+#if USE_NCURSES
+	endwin ();
+#endif
+
+	for (iter = ping_iterator_get (ping);
+			iter != NULL;
+			iter = ping_iterator_next (iter))
 	{
-		if (ts->tv_sec == 0)
+		ping_context_t *context;
+
+		context = ping_iterator_get_context (iter);
+
+		printf ("\n--- %s ping statistics ---\n"
+				"%i packets transmitted, %i received, %.2f%% packet loss, time %.1fms\n",
+				context->host, context->req_sent, context->req_rcvd,
+				context_get_packet_loss (context),
+				context->latency_total);
+
+		if (context->req_rcvd != 0)
 		{
-			ts->tv_nsec = 0;
-			return;
+			double average;
+			double deviation;
+
+			average = context_get_average (context);
+			deviation = context_get_stddev (context);
+
+			printf ("rtt min/avg/max/sdev = %.3f/%.3f/%.3f/%.3f ms\n",
+					context->latency_min,
+					average,
+					context->latency_max,
+					deviation);
 		}
 
-		ts->tv_sec  -= 1;
-		ts->tv_nsec += 1000000000;
+		ping_iterator_set_context (iter, NULL);
+		context_destroy (context);
 	}
 
-	while (ts->tv_nsec >= 1000000000)
-	{
-		ts->tv_sec  += 1;
-		ts->tv_nsec -= 1000000000;
-	}
-}
+	return (0);
+} /* }}} int post_loop_hook */
 
-static void time_calc (struct timespec *ts_dest,
-		const struct timespec *ts_int,
-		const struct timeval  *tv_begin,
-		const struct timeval  *tv_end)
-{
-	ts_dest->tv_sec = tv_begin->tv_sec + ts_int->tv_sec;
-	ts_dest->tv_nsec = (tv_begin->tv_usec * 1000) + ts_int->tv_nsec;
-	time_normalize (ts_dest);
-
-	/* Assure that `(begin + interval) > end'.
-	 * This may seem overly complicated, but `tv_sec' is of type `time_t'
-	 * which may be `unsigned. *sigh* */
-	if ((tv_end->tv_sec > ts_dest->tv_sec)
-			|| ((tv_end->tv_sec == ts_dest->tv_sec)
-				&& ((tv_end->tv_usec * 1000) > ts_dest->tv_nsec)))
-	{
-		ts_dest->tv_sec  = 0;
-		ts_dest->tv_nsec = 0;
-		return;
-	}
-
-	ts_dest->tv_sec = ts_dest->tv_sec - tv_end->tv_sec;
-	ts_dest->tv_nsec = ts_dest->tv_nsec - (tv_end->tv_usec * 1000);
-	time_normalize (ts_dest);
-}
-
-int main (int argc, char **argv)
+int main (int argc, char **argv) /* {{{ */
 {
 	pingobj_t      *ping;
 	pingobj_iter_t *iter;
@@ -475,6 +913,10 @@ int main (int argc, char **argv)
 				fprintf (stderr, "Adding host `%s' failed: %s\n", host, errmsg);
 				continue;
 			}
+			else
+			{
+				host_num++;
+			}
 		}
 
 #if _POSIX_SAVED_IDS
@@ -511,6 +953,10 @@ int main (int argc, char **argv)
 			fprintf (stderr, "Adding host `%s' failed: %s\n", argv[i], errmsg);
 			continue;
 		}
+		else
+		{
+			host_num++;
+		}
 	}
 
 	/* Permanently drop root privileges if we're setuid-root. */
@@ -526,32 +972,7 @@ int main (int argc, char **argv)
 	saved_set_uid = (uid_t) -1;
 #endif
 
-	i = 0;
-	for (iter = ping_iterator_get (ping);
-			iter != NULL;
-			iter = ping_iterator_next (iter))
-	{
-		ping_context_t *context;
-		size_t buffer_size;
-
-		context = context_create ();
-
-		buffer_size = sizeof (context->host);
-		ping_iterator_get_info (iter, PING_INFO_HOSTNAME, context->host, &buffer_size);
-
-		buffer_size = sizeof (context->addr);
-		ping_iterator_get_info (iter, PING_INFO_ADDRESS, context->addr, &buffer_size);
-
-		buffer_size = 0;
-		ping_iterator_get_info (iter, PING_INFO_DATA, NULL, &buffer_size);
-
-		printf ("PING %s (%s) %zu bytes of data.\n",
-				context->host, context->addr, buffer_size);
-
-		ping_iterator_set_context (iter, (void *) context);
-
-		i++;
-	}
+	ping_initialize_contexts (ping);
 
 	if (i == 0)
 		return (1);
@@ -564,8 +985,11 @@ int main (int argc, char **argv)
 		return (1);
 	}
 
+	pre_loop_hook (ping);
+
 	while (opt_count != 0)
 	{
+		int index;
 		int status;
 
 		if (gettimeofday (&tv_begin, NULL) < 0)
@@ -581,13 +1005,16 @@ int main (int argc, char **argv)
 			return (1);
 		}
 
+		index = 0;
 		for (iter = ping_iterator_get (ping);
 				iter != NULL;
 				iter = ping_iterator_next (iter))
 		{
-			print_host (iter);
+			update_host_hook (iter, index);
+			index++;
 		}
-		fflush (stdout);
+
+		pre_sleep_hook (ping);
 
 		/* Don't sleep in the last iteration */
 		if (opt_count == 1)
@@ -616,48 +1043,17 @@ int main (int argc, char **argv)
 			}
 		}
 
+		post_sleep_hook (ping);
+
 		if (opt_count > 0)
 			opt_count--;
 	} /* while (opt_count != 0) */
 
-	for (iter = ping_iterator_get (ping);
-			iter != NULL;
-			iter = ping_iterator_next (iter))
-	{
-		ping_context_t *context;
-
-		context = ping_iterator_get_context (iter);
-
-		printf ("\n--- %s ping statistics ---\n"
-				"%i packets transmitted, %i received, %.2f%% packet loss, time %.1fms\n",
-				context->host, context->req_sent, context->req_rcvd,
-				100.0 * (context->req_sent - context->req_rcvd) / ((double) context->req_sent),
-				context->latency_total);
-
-		if (context->req_rcvd != 0)
-		{
-			double num_total;
-			double average;
-			double deviation;
-
-			num_total = (double) context->req_rcvd;
-
-			average = context->latency_total / num_total;
-			deviation = sqrt (((num_total * context->latency_total_square) - (context->latency_total * context->latency_total))
-					/ (num_total * (num_total - 1.0)));
-
-			printf ("rtt min/avg/max/sdev = %.3f/%.3f/%.3f/%.3f ms\n",
-					context->latency_min,
-					average,
-					context->latency_max,
-					deviation);
-		}
-
-		ping_iterator_set_context (iter, NULL);
-		context_destroy (context);
-	}
+	post_loop_hook (ping);
 
 	ping_destroy (ping);
 
 	return (0);
-}
+} /* }}} int main */
+
+/* vim: set fdm=marker : */
