@@ -1,6 +1,6 @@
 /**
  * Object oriented C module to send ICMP and ICMPv6 `echo's.
- * Copyright (C) 2006  Florian octo Forster <octo at verplant.org>
+ * Copyright (C) 2006-2009  Florian octo Forster <octo at verplant.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,23 +24,17 @@
 # include <config.h>
 #endif
 
-#if HAVE_STDLIB_H
-# include <stdlib.h>
-#endif
-#if HAVE_UNISTD_H
-# include <unistd.h>
-#endif
-#if HAVE_SYS_TYPES_H
-# include <sys/types.h>
-#endif
-#if HAVE_SYS_SOCKET_H
-# include <sys/socket.h>
-#endif
+#include <stdlib.h>
+#include <unistd.h>
+#include <sys/types.h>
+#include <sys/socket.h>
 
 #ifdef __cplusplus
 extern "C" {
 #endif
 
+#define OPING_VERSION 1003003
+
 /*
  * Type definitions
  */
@@ -57,6 +51,7 @@ typedef struct pingobj pingobj_t;
 #define PING_OPT_AF      0x04
 #define PING_OPT_DATA    0x08
 #define PING_OPT_SOURCE  0x10
+#define PING_OPT_DEVICE  0x20
 
 #define PING_DEF_TIMEOUT 1.0
 #define PING_DEF_TTL     255
@@ -79,13 +74,16 @@ int ping_host_remove (pingobj_t *obj, const char *host);
 pingobj_iter_t *ping_iterator_get (pingobj_t *obj);
 pingobj_iter_t *ping_iterator_next (pingobj_iter_t *iter);
 
-#define PING_INFO_HOSTNAME 1
-#define PING_INFO_ADDRESS  2
-#define PING_INFO_FAMILY   3
-#define PING_INFO_LATENCY  4
-#define PING_INFO_SEQUENCE 5
-#define PING_INFO_IDENT    6
-#define PING_INFO_DATA     7
+#define PING_INFO_HOSTNAME  1
+#define PING_INFO_ADDRESS   2
+#define PING_INFO_FAMILY    3
+#define PING_INFO_LATENCY   4
+#define PING_INFO_SEQUENCE  5
+#define PING_INFO_IDENT     6
+#define PING_INFO_DATA      7
+#define PING_INFO_USERNAME  8
+#define PING_INFO_DROPPED   9
+#define PING_INFO_RECV_TTL 10
 int ping_iterator_get_info (pingobj_iter_t *iter, int info,
 		void *buffer, size_t *buffer_len);
 
